# -*- coding: utf-8 -*-
################################################################################
# pcollections/test/_dict.py
# Declaration of tests for the pdict and tdict types.
# By Noah C. Benson

from random import randint
from unittest import TestCase

from .._dict import (pdict, tdict)
from .._lazy import (lazy, ldict)

class TestPDict(TestCase):
    """Tests for the `pdict` and `tdict` classes.

    This both runs a number of simple tests of the `pdict` API and a series of
    randomized tests in which `pdict` must match the behavior of Python's
    native `dict` type.
    """
    def test_api(self):
        """Tests/demonstrates the basic pdict API."""
        # An empty pdict can be created with `pdict()`.
        e = pdict()
        self.assertEqual(len(e), 0)
        self.assertIsInstance(e, pdict)
        # This is identical to the pdict.empty object.
        self.assertIs(e, pdict.empty)
        # A pdict can also be created by passing an iterable to the type.
        p1 = pdict(zip(range(0,10), range(10,20)))
        self.assertEqual(len(p1), 10)
        # Element access is like with the dict type.
        self.assertEqual(p1[0], 10)
        self.assertEqual(p1[5], 15)
        self.assertEqual(p1[9], 19)
        # pdicts can be equal to each other and to other Mappings like dicts.
        l1 = dict(zip(range(0,10), range(10,20)))
        self.assertEqual(l1, p1)
        self.assertEqual(p1, l1)
        self.assertEqual(p1, pdict(l1.items()))
        self.assertNotEqual(p1, p1.keys())
        self.assertNotEqual(pdict.empty, [])
        self.assertNotEqual(pdict.empty, '')
        # Clearing a pdict always yields the empty pdict.
        self.assertIs(p1.clear(), pdict.empty)
        # Copying a pdict always just returns the pdict (it is immutable).
        self.assertIs(p1.copy(), p1)
        # pdicts are hashable.
        self.assertIsInstance(hash(p1), int)
        # pdicts contain their keyss.
        for k in range(0,10):
            self.assertIn(k, p1)
        self.assertNotIn(-1, p1)
        self.assertNotIn(11, p1)
        # They can be iterated.
        self.assertEqual(l1.keys(), set(iter(p1)))
        # They have lengths.
        self.assertEqual(len(p1), 10)
        # Mutating elements is done by creating a copy with the desired mutation
        # using the set method. This leaves the original object unchanged.
        p2 = p1.set(0, 100)
        self.assertEqual(p2[0], 100)
        self.assertEqual(p1[0], 10)
        self.assertNotEqual(p1, p2)
        p2 = p1.set(10, 100)
        self.assertEqual(p2[10], 100)
        self.assertNotIn(10, p1)
        self.assertNotEqual(p1, p2)
        # Deleting elements can be done using the delete or drop methods.
        self.assertEqual(p2.delete(10), p1)
        self.assertEqual(p2.drop(10), p1)
        # Drop and delete only differ in their behavior when the index is
        # invalid.
        with self.assertRaises(KeyError):
            p2.delete(20)
        self.assertEqual(p2, p2.drop(20))
        # The update method returns a merged dictionary of the arguments.
        l2 = dict(p2)
        ext = {4:40,5:50,6:60}
        l2.update(ext)
        self.assertEqual(p2.update({4:40,5:50,6:60}), l2)
        l2 = dict(p2)
        l2.update(a=10, b=20)
        self.assertEqual(p2.update(a=10, b=20), l2)
        # The pop method may be used to extract an item.
        self.assertEqual(p2.pop(1), (11, p2.drop(1)))
        # There are also batch methods for most of the operations.
        self.assertEqual(pdict.empty.setall(range(3), range(0,30,10)),
                         {0:0, 1:10, 2:20})
    def test_immutable(self):
        """Ensures that `pdict` throws the right errors when one mutates it."""
        l = pdict(zip(range(10), range(0,100,10)))
        # Cannot set-item.
        with self.assertRaises(TypeError):
            l[0] = 10
        # Cannot del-item.
        with self.assertRaises(TypeError):
            del l[0]
        # Cannot set-attr.
        with self.assertRaises(TypeError):
            l._top = -10
    def test_random(self):
        "Performs a randomized test on the pdict type."
        nops = 100
        valmax = 1000
        keymax = 200
        p = pdict()
        t = tdict()
        l = dict()
        for opnum in range(nops):
            op = randint(0, 9)
            key = randint(0, keymax)
            el = randint(0, valmax)
            if op < 4:
                ii = randint(0, len(l))
                #print(f"set ({ii}, {el}) at {len(l)}")
                p = p.set(ii, el)
                t[ii] = el
                l[ii] = el
            elif op == 4:
                if len(p) > 0:
                    k = list(l.keys())[randint(0, len(l) - 1)]
                    lel = l.pop(k)
                    #print(f"pop {k} {lel} at {len(l)}")
                    (pel, p) = p.pop(k)
                    tel = t.pop(k)
                    self.assertEqual(pel, tel)
                    self.assertEqual(pel, lel)
            elif op == 5:
                if len(p) > 0:
                    k = list(l.keys())[randint(0, len(l) - 1)]
                    #print(f"drop {k} at {len(l)}")
                    p = p.drop(k)
                    del t[k]
                    del l[k]
            #print('   -', repr(p), ' [', p._start, ']')
            #print('   -', repr(t), ' [', t._start, ']')
            #print('   -', repr(l))
            self.assertEqual(p, t)
            self.assertEqual(p, l)
            self.assertEqual(l, t)
            if randint(0, 20) == 0:
                tmp = t.persistent()
                self.assertEqual(tmp, p)
                self.assertEqual(tmp, t)

class TestLDict(TestCase):
    """Tests for the `ldict` and `lazy` classes.

    Because the `ldict` code is based largely on the `pdict` code, only API
    tests (and not randomized tests) are run for the lazy dictionary type.
    """
    def test_api(self):
        """Tests/demonstrates the basic ldict API."""
        # An empty ldict can be created with `ldict()`.
        e = ldict()
        self.assertEqual(len(e), 0)
        self.assertIsInstance(e, ldict)
        # An ldict is also a pdict.
        self.assertIsInstance(e, pdict)
        # This is identical to the ldict.empty object.
        self.assertIs(e, ldict.empty)
        # ... but not to pdict.empty.
        self.assertIsNot(e, pdict.empty)
        # An ldict can also be created by passing an iterable to the type.
        p1 = ldict(zip(range(0,10), range(10,20)))
        self.assertEqual(len(p1), 10)
        # Element access is like with the dict type.
        self.assertEqual(p1[0], 10)
        self.assertEqual(p1[5], 15)
        self.assertEqual(p1[9], 19)
        self.assertEqual(p1.get(0), 10)
        self.assertEqual(p1.get(5), 15)
        self.assertEqual(p1.get(100), None)
        self.assertEqual(p1.get(100, -1), -1)
        # ldicts can be equal to each other and to other Mappings like dicts.
        l1 = dict(zip(range(0,10), range(10,20)))
        self.assertEqual(l1, p1)
        self.assertEqual(p1, l1)
        self.assertEqual(p1, pdict(l1.items()))
        self.assertEqual(p1, ldict(l1.items()))
        self.assertNotEqual(p1, p1.keys())
        self.assertNotEqual(ldict.empty, [])
        self.assertNotEqual(ldict.empty, '')
        # Clearing an ldict always yields the empty ldict.
        self.assertIs(p1.clear(), ldict.empty)
        # Copying an ldict always just returns the ldict (it is immutable).
        self.assertIs(p1.copy(), p1)
        # ldicts are hashable (assuming all values are hashable).
        self.assertIsInstance(hash(p1), int)
        # ldicts contain their keyss.
        for k in range(0,10):
            self.assertIn(k, p1)
        self.assertNotIn(-1, p1)
        self.assertNotIn(11, p1)
        # They can be iterated.
        self.assertEqual(l1.keys(), set(iter(p1)))
        # They have lengths.
        self.assertEqual(len(p1), 10)
        # Mutating elements is done by creating a copy with the desired mutation
        # using the set method. This leaves the original object unchanged.
        p2 = p1.set(0, 100)
        self.assertEqual(p2[0], 100)
        self.assertEqual(p1[0], 10)
        self.assertNotEqual(p1, p2)
        p2 = p1.set(10, 100)
        self.assertEqual(p2[10], 100)
        self.assertNotIn(10, p1)
        self.assertNotEqual(p1, p2)
        # Deleting elements can be done using the delete or drop methods.
        self.assertEqual(p2.delete(10), p1)
        self.assertEqual(p2.drop(10), p1)
        # Drop and delete only differ in their behavior when the index is
        # invalid.
        with self.assertRaises(KeyError):
            p2.delete(20)
        self.assertEqual(p2, p2.drop(20))
        # The update method returns a merged dictionary of the arguments.
        l2 = dict(p2)
        ext = {4:40,5:50,6:60}
        l2.update(ext)
        self.assertEqual(p2.update({4:40,5:50,6:60}), l2)
        l2 = dict(p2)
        l2.update(a=10, b=20)
        self.assertEqual(p2.update(a=10, b=20), l2)
        # The pop method may be used to extract an item.
        self.assertEqual(p2.pop(1), (11, p2.drop(1)))
        # There are also batch methods for most of the operations.
        self.assertEqual(ldict.empty.setall(range(3), range(0,30,10)),
                         {0:0, 1:10, 2:20})
    def test_lazy(self):
        "Tests the lazy aspects of the ldict class."
        # Lazy dictionaries don't evaluate lazy arguments until they are
        # requested, and they only evaluate them once.
        def counter(n):
            counter.count += n
            return counter.count
        counter.count = 0
        p1 = ldict(a=lazy(counter, 1), b=lazy(counter, 10))
        self.assertEqual(counter.count, 0)
        self.assertEqual(p1['a'], 1)
        self.assertEqual(counter.count, 1)
        # Reaccessing the key a will not call counter() again.
        self.assertEqual(p1['a'], 1)
        self.assertEqual(counter.count, 1)
        # However, the key b has not yet been evaluated, so accessing it will
        # call counter().
        self.assertEqual(p1['b'], 11)
        self.assertEqual(counter.count, 11)
        self.assertEqual(p1['b'], 11)
        self.assertEqual(counter.count, 11)
        # Iteration also reifies the values.
        counter.count = 0
        p1 = ldict(a=lazy(counter, 1), b=lazy(counter, 10))
        self.assertEqual(list(p1.values()), [1, 11])
        self.assertEqual(p1['a'], 1)
        self.assertEqual(p1['b'], 11)
        self.assertEqual(counter.count, 11)
        # Converstion into a tdict respects the lazy items.
        counter.count = 0
        p1 = ldict(a=lazy(counter, 1), b=lazy(counter, 10))
        t1 = p1.transient()
        self.assertEqual(t1['a'], 1)
        self.assertEqual(t1['b'], 11)
        # The to_pdict method can be used to reveeal lazy elements.
        counter.count = 0
        p1 = ldict(a=lazy(counter, 1), b=lazy(counter, 10))
        t1 = p1.to_pdict()
        self.assertIsInstance(t1['a'], lazy)
        self.assertIsInstance(t1['b'], lazy)
        # Conversion back into a lazy dict will hide these.
        p2 = ldict(t1)
        self.assertEqual(p2['a'], 1)
        self.assertEqual(p2['b'], 11)
        # Note that the new dictionary in this case shares the lazy values with
        # the old dictionary.
        self.assertEqual(p1['a'], 1)
        self.assertEqual(p1['b'], 11)
        # Equality comparisons depend on the reified, not lazy, values.
        counter.count = 0
        p1 = ldict(a=lazy(counter, 1), b=lazy(counter, 10))
        self.assertEqual(p1, {'a': 1, 'b': 11})
    def test_immutable(self):
        """Ensures that `pdict` throws the right errors when one mutates it."""
        l = ldict(zip(range(10), range(0,100,10)))
        # Cannot set-item.
        with self.assertRaises(TypeError):
            l[0] = 10
        # Cannot del-item.
        with self.assertRaises(TypeError):
            del l[0]
        # Cannot set-attr.
        with self.assertRaises(TypeError):
            l._top = -10



# -*- coding: utf-8 -*-
################################################################################
# pcollections/_lazy.py
# The lazy dictionary and list implementations.
# By Noah C. Benson

from functools import partial
from threading import RLock

from phamt import (PHAMT, THAMT)

from ._list import (
    plist,
    tlist
)
from ._dict import (
    pdict_items,
    pdict_values,
    pdict,
    tdict,
    tdict_items,
    tdict_values
)


#===============================================================================
# Lazy Value Type

class LazyError(RuntimeError):
    """A runtime error that occurs while evaluating a lazy value.

    See also: `lazy`.
    """
    __slots__ = 'partial'
    def __init__(self, partial):
        self.partial = partial
    def __str__(self):
        (fn, args, kwargs) = self.partial
        fnname = getattr(fn, '__name__', '<anonymous>')
        errmsg = ('lazy raised error during call to '
                  f'{fnname}{tuple(args)}')
        if len(kwargs) > 0:
            opts = ', '.join([f'{k}={v}' for (k,v) in kwargs.items()])
            errmsg = f'{errmsg[:-1]}, {opts})'
        return errmsg
    def __repr__(self):
        return str(self)
class lazy:
    """A callable like `partial` for lazily-computed values.

    The `lazy` type is constructed identically to the `partial` type. Unlike
    `partial`, however, `lazy` values must have their entire argument lists
    instantiated at the time of construction--i.e., it is not possible to call
    a lazy partial with additional arguments.

    `l = lazy(fn, *args, **kwargs)` stores the given callable `fn` with the
    given `args` and `kwargs` as arguments. When the lazy value of `l` is later
    requested (via `l()`), the value is cached, and the partial data is
    forgotten.
    """
    __slots__ = ('partial', 'value')
    def __new__(cls, fn, *args, **kw):
        # Make sure the function is callable.
        if not callable(fn):
            raise TypeError(f"lazy({fn}) must be given a callable function")
        # Allocate an object.
        obj = object.__new__(cls)
        # Create the partial object.
        part = partial(fn, *args, **kw)
        # We want to prepare an error to raise if something happens during the
        # calculation of the lazy value in the __call__ method.
        try:
            raise LazyError(part)
        except LazyError as e:
            error = e
        # Set the appropriate members.
        object.__setattr__(obj, 'partial', (part, RLock(), error))
        # We set value to an RLock for use in the calculation.
        object.__setattr__(obj, 'value', None)
        # That's all that is needed.
        return obj
    def __call__(self):
        part = self.partial
        if part is None:
            return self.value
        else:
            (part, rlock, init_error) = part
            # Acquire the rlock then re-check that the lazy hasn't already been
            # calculated (at which point self.partial will be None).
            rlock.acquire()
            try:
                if self.partial is None:
                    val = self.value
                else:
                    val = part()
                    # We've successfully calculated the value; set the members
                    # appropriately.
                    object.__setattr__(self, 'value', val)
                    object.__setattr__(self, 'partial', None)
                return val
            except Exception as partial_eror:
                # If an exception occurs, we want to raise an error that traces
                # back to the initialization of this object.
                raise init_error
            finally:
                rlock.release()
    def __repr__(self):
        s = 'ready' if self.is_cached() else 'waiting'
        return f"lazy(<{id(self)}>: {s})"
    def __str__(self):
        return f"lazy(<{id(self)}>)"
    def is_ready(self):
        """Returns `True` if the lazy value is cached and `False` otherwise.

        Returns
        -------
        boolean
            `True` if the given lazy object has cached its value and `False`
            otherwise.
        """
        return self.partial is None
def unlazy(obj):
    """Returns the cached value of a lazy object or the object if not lazy."""
    if isinstance(obj, lazy):
       return obj()
    else:
       return obj


#===============================================================================
# The Lazy List Type

class llist(plist):
    """A persistent lazy list type.

    The `llist` type is identical to the `plist` type, with the exception that
    any `lazy` value contained in an `llist` is dereferenced prior to being
    returned by item extraction from the list or iteration. Essentially, an
    `llist` containing a lazy value behaves equivalently to a `plist` whose
    values were all precalculated. Note that hashing an `llist` object results
    in all lazy values being calculated.

    The `llist.transient` method returns a normal `tlist`, but one the lazy
    values are not dereferenced, meaning that laziness is respected for
    edits. However, if one requests a lazy value from the transient list, the
    `lazy` object itself will be returned. In order to recreate a lazy list
    (instead of a persistent list) from the transient list, one must use
    `llist(t)` instead of `t.persistent()`; the latter call will return a
    `plist` containing `lazy` objects.
    """
    empty = None
    __slots__ = ()
    def __iter__(self):
        it = plist.__iter__(self)
        return map(lambda u: u() if isinstance(u, lazy) else u, it)
    def __getitem__(self, k):
        el = plist.__getitem__(self, k)
        return el() if isinstance(el, lazy) else el
    def is_lazy(self, index):
        """Determines if the given key is mapped to a `lazy` value.

        `is_lazy` determines whether the associated key is mapped to a `lazy`
        object, but it does not determine if the lazy value is cached. To query
        whether a key is mapped to a value that is uncached, see the `is_ready`
        method.
        """
        v = plist.__getitem__(self, index)
        return isinstance(v, lazy)
    def is_ready(self, index):
        """Determines if the given key's value can be immediately returned.

        `is_ready` determines whether the associated key is either mapped to a
        non-`lazy` value or is mapped to a `lazy` value that is cached.
        """
        v = plist.__getitem__(self, index)
        if isinstance(v, lazy):
            return v.is_ready()
        else:
            return True
    def ready_all(self):
        "Caches all lazy items then returns the list."
        for el in self._phamt:
            if isinstance(el, lazy):
                el()
        return self
    def to_plist(self):
        """Returns a `plist`  of the lazy list with `lazy` values uncached.

        Whereas `plist(l)` will return a `plist` whose values are all the cached
        values of the `llist` `l`, the method `l.to_plist()` returns a copy of
        `l` where indices of `l` that lazily compute their values are mapped to
        their associated `lazy` objects. This is essentially a way to expose the
        raw values of a lazy list.
        """
        return plist._new(self._phamt, self._start)
    def getlazy(self, index):
        """Like getitem, but returns lazy objects instead of their results.

        For an `llist` variable `l`, `l.getlazy(ii)` is equivalent to `l[ii]`
        except that if the index `ii` maps to a lazy value, then `l.getlazy`
        will return the lazy object rather than evaluating it and returning the
        reified value.
        """
        return plist.__getitem__(self, index)
    def clear(self):
        return llist.empty
    def transient(self):
        return tllist._new(THAMT(self._phamt), self._start, self)
# Setup the llist.empty static member.
llist.empty = llist._new(PHAMT.empty, 0)

# The Transient Lazy List Type -------------------------------------------------
class tllist(tlist):
    """A transient lazy list type.

    Transient lazy lists are like transient lists (`tlist`) with the only
    difference being that they automatically return the reified values of lazy
    elements instead of the `lazy` objects themselves.
    """
    __slots__ = ()
    def __iter__(self):
        return map(unlazy, tlist.__iter__(self))
    def persistent(self):
        """Efficiently copies the tllist into an llist and returns the llist."""
        if len(self._thamt) == 0:
            return llist.empty
        elif self._orig is None:
            return llist._new(self._thamt.persistent(), self._start)
        else:
            return self._orig
    def __getitem__(self, k):
        return unlazy(tlist.__getitem__(self, k))
    def getlazy(self, k):
        """Returns an element of the list without dereferecing lazy elements."""
        return tlist.__getitem__(self, k)
        


#===============================================================================
# The Lazy Dictionary Type

class ldict_items(pdict_items):
    __slots__ = ()
    def _from_kv(self, kv):
        (k,v) = kv
        if isinstance(v, lazy):
            return (k, v())
        else:
            return kv
class ldict_values(pdict_values):
    __slots__ = ()
    def _from_kv(self, kv):
        v = kv[1]
        if isinstance(v, lazy):
            v = v()
        return v
class ldict(pdict):
    """A persistent lazy dict type.

    The `ldict` type is identical to the `pdict` type, with the exception that
    any `lazy` value contained in an `ldict` is dereferenced prior to being
    returned by item extraction from the dict or iteration. Essentially, an
    `ldict` containing a lazy value behaves equivalently to a `pdict` whose
    values were all precalculated. Note that hashing an `ldict` object results
    in all lazy values being calculated.

    The `ldict.transient` method returns a normal `tdict`, but one the lazy
    values are not dereferenced, meaning that laziness is respected for
    edits. However, if one requests a lazy value from the transient dict, the
    `lazy` object itself will be returned. In order to recreate a lazy dict
    (instead of a persistent dict) from the transient dict, one must use
    `ldict(t)` instead of `t.persistent()`; the latter call will return a
    `pdict` containing `lazy` objects.
    """
    empty = None
    __slots__ = ()
    def __getitem__(self, key):
        v = pdict.__getitem__(self, key)
        if isinstance(v, lazy):
            v = v()
        return v
    def get(self, key, default=None):
        v = pdict.get(self, key, default)
        if isinstance(v, lazy):
            v = v()
        return v
    def items(self):
        return ldict_items(self)
    def values(self):
        return ldict_values(self)
    def is_lazy(self, index):
        """Determines if the given key is mapped to a `lazy` value.

        `is_lazy` determines whether the associated key is mapped to a `lazy`
        object, but it does not determine if the lazy value is cached. To query
        whether a key is mapped to a value that is uncached, see the `is_ready`
        method.
        """
        v = pdict.__getitem__(self, index)
        return isinstance(v, lazy)
    def is_ready(self, index):
        """Determines if the given key's value can be immediately returned.

        `is_ready` determines whether the associated key is either mapped to a
        non-`lazy` value or is mapped to a `lazy` value that is cached.
        """
        v = pdict.__getitem__(self, index)
        if isinstance(v, lazy):
            return v.is_ready()
        else:
            return True
    def ready_all(self):
        "Caches all lazy items then returns the dictionary."
        for arg in self._els:
            (k,v) = arg[1][0]
            if isinstance(v, lazy):
                v()
        return self
    def to_pdict(self):
        """Returns a `pdict`  of the lazy dict with `lazy` values uncached.

        Whereas `pdict(ld)` will return a `pdict` whose values are all the
        cached values of the `ldict` `ld`, the method `ld.to_pdict()` returns a
        copy of `ld` where keys of `ld` that lazily compute their values are
        mapped to their associated `lazy` objects. This is essentially a way to
        expose the raw values of a lazy dictionary.
        """
        return pdict._new(self._els, self._idx, self._top)
    def getlazy(self, key, default=None):
        """Like get, but returns lazy objects instead of their results.

        For an `ldict` variable `d`, `d.getlazy(k, default)` is equivalent to
        `d.get(k, default)` except that if the key `k` maps to a lazy value,
        then `d.getlazy` will return the lazy object rather than evaluating it
        and returning the reified value.
        """
        return pdict.get(self, key, default)
    def clear(self):
        return ldict.empty
    def transient(self):
        return tldict._new(THAMT(self._els), THAMT(self._idx), self._top, self)
# Make the empty pdict.
ldict.empty = ldict._new(PHAMT.empty, PHAMT.empty, 0)

# The Transient Lazy Dictionary Type -------------------------------------------
class tldict_items(tdict_items):
    __slots__ = ()
    def _from_kv(self, kv):
        if isinstance(kv[1], lazy):
            return (kv[0], kv[1]())
        else:
            return kv
class tldict_values(tdict_values):
    __slots__ = ()
    def _from_kv(self, arg):
        return unlazy(arg[1])
    def __contains__(self, v):
        for (kv,_) in self._els:
            if not isinstance(kv[1], lazy):
                if kv[1] == v:
                    return True
        for (kv,_) in self._els:
            if isinstance(kv[1], lazy):
                if kv[1]() == v:
                    return True
        return False
class tldict(tdict):
    """A transient lazy dict type.

    Transient lazy dicts are like transient dicts (`tdict`) with the only
    difference being that they automatically return the reified values of lazy
    keys instead of the `lazy` objects themselves.
    """
    __slots__ = ()
    def persistent(self):
        """Efficiently copies the tllist into an llist and returns the llist."""
        if len(self._thamt) == 0:
            return ldict.empty
        elif self._orig is None:
            return ldict._new(self._thamt.persistent(), self._start)
        else:
            return self._orig
    def __getitem__(self, k):
        return unlazy(tdict.__getitem__(self, k))
    def get(self, k, default=None):
        return unlazy(tdict.get(self, k, default))
    def pop(self, *args):
        return unlazy(self.pop(*args))
    def items(self):
        return tldict_items(self)
    def values(self):
        return tldict_values(self)



# -*- coding: utf-8 -*-
################################################################################
# pcollections/abc/_map.py
# The definitions of the abstract base classes for the pcollections dict types.
# By Noah C. Benson

from collections.abc import (Mapping, MutableMapping)

from ._core import (Persistent, Transient)
from ..util import (seqstr)


#===============================================================================
# PersistentMapping

class PersistentMapping(Mapping, Persistent):
    """All the operations on a persistent mapping.

    Persistent mappings are mappings (i.e., objects that inherit from
    `collections.abc.Mapping`), but they differ from other mappings in that they
    support efficient updating by means of efficiently producing copies of
    themselves that incorporate requested changes.

    The following abstract methods must be implemented; if these methods are
    inherited from a superclass of `PersistentMapping`, that class is noted in
    parentheses.
     * `__len__` (`Sized`)
     * `__iter__` (`Iterable`)
     * `__getitem__` (`Mapping`)
     * `transient()` (`Persistent`)
     * `set(key, value)`
     * `drop(key)`
     * `clear()`

    Additionally, `PersistentMapping` includes default implementations of the
    following methods, which may or may not be optimal for any particular
    base-class.
     * `__setattr__` (`object`; raises a `TypeError`)
     * `__setitem__` (`object`; raises a `TypeError`)
     * `__str__` (`object`)
     * `__repr__` (`object`)
     * `__eq__` (`object`)
     * `__ne__` (`object`)
     * `__hash__` (`Hashable`)
     * `__contains__` (`Container`)
     * `get(key, default=None)` (`Mapping`)
     * `keys()` (`Mapping`)
     * `items()` (`Mapping`)
     * `values()` (`Mapping`)
     * `setdefault(key, default=None)` (`MutableMapping`)
     * `popitem()` (`MutableMapping`)
     * `pop(key)` (`MutableMapping`)
     * `copy()` (`Persistent`)
     * `delete(key)`
     * `remove(value)`
     * `setall(values)`
     * `dropall(key)`
     * `deleteall(key)`
     * `discardall(values)`
     * `removeall(values)`
     * `update(map, **kw)`
     * `__reduce__` (for pickling)
     * `__json__` (for `json_fix` module)
    """
    # Methods which must be implemented in the children.
    def set(self, key, val):
        """Returns a copy of the pdict that maps the given key to the given
        value."""
        raise NotImplementedError()
    def drop(self, key, error=False):
        """Returns a copy of the persistent mapping that does not include the
        given key.

        If the key is not in the dict, `drop` returns the original persistent
        mapping.
        """
        raise NotImplementedError()
    def clear(self):
        """Returns the empty persistent mapping of the same type."""
        raise NotImplementedError()
    # Methods that are probably fine for any child-class of PersistentMapping.
    def __str__(self):
        # We have a max length of 60 characters, not counting the delimiters.
        return f"{{|{seqstr(self, maxlen=60)}|}}"
    def __repr__(self):
        return f"{{|{seqstr(self)}|}}"
    def __hash__(self):
        return hash(frozenset(map(lambda u: u[1][0], self._els))) + 2
    def __contains__(self, k):
        try:
            self[k]
            return True
        except KeyError:
            return False
    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default
    def delete(self, key):
        """Returns a copy of the mersistent mapping the excludes the given key.

        If the key is not found in the mapping, a `KeyError` is raised.
        """
        if key in self:
            return self.drop(key)
        else:
            raise KeyError(key)
    def setall(self, keys, vals):
        """Returns a copy of the persistent mapping that includes all the object
        in the given iterables of keys and values.
        """
        t = self.transient()
        for (k,v) in zip(keys, vals):
            t[k] = v
        return type(self)(t)
    def dropall(self, keys):
        """Returns a copy of the persistent mapping that excludes all the keys
        in the given iterable.

        Keys that are not found in the mapping are ignored.
        """
        t = self.transient()
        for k in keys:
            t.drop(k)
        return type(self)(t)
    def deleteall(self, keys):
        """Returns a copy of the persistent mapping that excludes all the keys
        in the given iterable.

        If any key is not found in the mapping, then a KeyError error is raised.
        """
        t = self.transient()
        for k in keys:
            del t[k]
        return type(self)(t)
    def setdefault(self, key, default=None):
        """Returns a copy of the persistent mapping with the key inserted with a
        value of default, if key is not already in the mapping.
        """
        return self if key in self else self.set(key, default)
    def popitem(self):
        """Returns a 2-tuple whose first element is itself a 2-tuple, `(key,
        value)` from the persistent mapping, and whose second element is a copy
        of the mapping with the key-value pair removed.

        Raises `KeyError` if the persistent mapping is empty.
        """
        if len(self) == 0:
            raise KeyError("popitem(): persistent mapping is empty")
        kv = next(iter(self.items()))
        return (kv, self.drop(kv[0]))
    def pop(self, key, *args):
        """Returns a tuple of the value mapped to the given key and a copy of
        the persistent mapping with that key removed. 

        If the key is not found, the second argument is returned, if given,
        otherwise, a `KeyError` is raised.
        """
        nargs = len(args)
        if nargs > 1:
            raise TypeError(f"pop expected at most 2 arguments, got {nargs}")
        try:
            val = self[key]
            return (val, self.drop(key))
        except KeyError:
            # It's not here!
            if nargs == 0:
                raise
            else:
                return (args[0], self)
    def update(self, *args, **kw):
        """Returns a copy of the persistent mapping with the key-value pairs in
        the iterable/mapping `arg` and the keyword arguments included.

        Multiple arguments may be provided, but they must each be a mapping or
        an iterable of items.
        """
        t = self.transient()
        for arg in args:
            for (k,v) in (arg.items() if isinstance(arg, Mapping) else arg):
                t[k] = v
        for (k,v) in kw.items():
            t[k] = v
        return type(self)(t)
    def __reduce__(self):
        return (self.__new__, (type(self), list(self.items()),))
    def __json__(self):
        from json import dumps
        return dumps(dict(self))


#===============================================================================
# TransientMapping

class TransientMapping(MutableMapping, Transient):
    """All the operations on a transient mapping.

    Transient mappings are mutable mappings (i.e., objects that inherit from
    `collections.abc.MutableMapping`), but they differ from other mutable
    mappings in that they support efficient conversion to and from persistent
    mappings.

    The following abstract methods must be implemented; if these methods are
    inherited from a superclass of `TransientMapping`, that class is noted in
    parentheses.
     * `__len__` (`Sized`)
     * `__iter__` (`Iterable`)
     * `__getitem__` (`Mapping`)
     * `__setitem__` (`MutableMapping`)
     * `__delitem__` (`MutableMapping`)
     * `transient()` (`Persistent`)
     * `clear()`

    Additionally, `TransientMapping` includes default implementations of the
    following methods, which may or may not be optimal for any particular
    base-class.
     * `__setattr__` (`object`; raises a `TypeError`)
     * `__str__` (`object`)
     * `__repr__` (`object`)
     * `__eq__` (`object`)
     * `__ne__` (`object`)
     * `__contains__` (`Container`)
     * `get(key, default=None)` (`Mapping`)
     * `keys()` (`Mapping`)
     * `items()` (`Mapping`)
     * `values()` (`Mapping`)
     * `setdefault(key, default=None)` (`MutableMapping`)
     * `popitem()` (`MutableMapping`)
     * `pop()` (`MutableMapping`)
     * `update(map, **kw)` (`MutableMapping`)
     * `copy()` (`Transient`)
     * `__reduce__` (for pickling)
     * `__json__` (for the `json_fix` module)
    """
    # Methods which must be implemented in the children.
    def clear(self):
        """Returns the empty persistent mapping of the same type."""
        raise NotImplementedError()
    # Methods that are probably fine for any child-class of PersistentMapping.
    def __str__(self):
        # We have a max length of 60 characters, not counting the delimiters.
        return f"{{<{seqstr(self, maxlen=60)}>}}"
    def __repr__(self):
        #s = repr(dict(self))
        #return f"{{<{s[1:-1]}>}}"
        return f"{{|{seqstr(self)}|}}"
    def __contains__(self, k):
        try:
            self[k]
            return True
        except KeyError:
            return False
    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default
    def setdefault(self, key, default=None):
        """Insert key with a value of default if key is not in the dictionary.

        Return the value for the key if key is in the dictionary, else default.
        """
        try:
            return self[key]
        except KeyError:
            pass
        self[key] = default
        return default
    def popitem(self):
        """Remove and return a (key, value) pair as a 2-tuple.

        Raises `KeyError` if the persistent mapping is empty.
        """
        if len(self) == 0:
            raise KeyError("popitem(): transient mapping is empty")
        kv = next(iter(self.items()))
        del self[kv[0]]
        return kv
    def pop(self, key, *args):
        """Removes the specified key and returns the corresponding value.

        If the key is not found, the second argument is returned, if given,
        otherwise, a `KeyError` is raised.
        """
        nargs = len(args)
        if nargs > 1:
            raise TypeError(f"pop expected at most 2 arguments, got {nargs}")
        try:
            val = self[key]
            del self[key]
            return val
        except KeyError:
            # It's not here!
            if nargs == 0:
                raise
            else:
                return args[0]
    def update(self, *args, **kw):
        """Updates the transient mapping with items from the arguments and
        options.

        Multiple arguments may be provided, but they must each be either a
        mapping or an iterable of items.
        """
        for arg in args:
            for (k,v) in (arg.items() if isinstance(arg, Mapping) else arg):
                self[k] = v
        for (k,v) in kw.items():
            self[k] = v
    def copy(self):
        """Returns a copy of the transient mapping."""
        return self.persistent().transient()
    # The below handle pickling cases.
    def __reduce__(self):
        return (self.__new__, (type(self), list(self.items()),))
    def __json__(self):
        from json import dumps
        return dumps(dict(self))
        



# -*- coding: utf-8 -*-
################################################################################
# pcollections/abc/_seq.py
# The definitions of the abstract base classes for the pcollections sequence
# types.
# By Noah C. Benson

from numbers         import (Integral)
from collections.abc import (Sequence, MutableSequence)

from ._core import (Persistent, Transient)
from ..util import (seqstr, seqcmp)


#===============================================================================
# PersistentSequence

class PersistentSequence(Persistent, Sequence):
    """All the operations on a persistent sequence.

    Persistent sequences are sequences (i.e., objects that inherit from
    `collections.abc.Sequence`), but they differ from other sequences in that
    they support efficient updating by means of efficiently producing copies of
    themselves that incorporate requested changes.

    The following abstract methods must be implemented; if these methods are
    inherited from a superclass of `PersistentSequence`, that class is noted in
    parentheses.
     * `set(index, object)`
     * `delete(index=-1)`
     * `append(object)`
     * `prepend(object)`
     * `insert(index, object)`
     * `clear()`
     * `transient()` (`Persistent`)
     * `__iter__` (`Iterable`)
     * `__len__` (`Sized`)
     * `__getitem__` (`Sequence`)
     * `__reduce__` (for pickling)

    Additionally, `PersistentSequence` includes default implementations of the
    following methods, which may or may not be optimal for any particular
    base-class.
     * `__str__` (`object`)
     * `__repr__` (`object`)
     * `__eq__` (`object`)
     * `__hash__` (`Hashable`)
     * `__contains__` (`Container`)
     * `__reversed__` (`Reversible`)
     * `count` (`Sequence`)
     * `extend(iterable)` (`Sequence`)
     * `index(value)` (`Sequence`)
     * `copy()` (`Persistent`)
     * `__add__`
     * `__radd__`
     * `__mul__`
     * __rmul__`
     * `drop(index=-1)`
     * `pop(index=-1)`
     * `remove(value)`
     * `sort()`
     * `reverse()`
     * `__json__` (for `json_fix` module)
    """
    # Methods which must be implemented in the children.
    def set(self, index, obj):
        """Returns a copy of the persistent sequence with the given index set to
        the given object.
        """
        raise NotImplementedError()
    def delete(self, index=-1):
        """"Returns a copy of the persistent sequence with the item at index
        removed (default index: last)."""
        raise NotImplementedError()
    def append(self, obj):
        """Returns a copy of the persistent sequence with object appended."""
        raise NotImplementedError()
    def prepend(self, obj):
        """Returns a copy of the persistent sequence with object prepended."""
        raise NotImplementedError()
    def insert(self, index, obj):
        "Returns a new persistent sequence with object inserted before index."
        raise NotImplementedError()
    def clear(self):
        """Returns an empty persistent sequence of the same type."""
        raise NotImplementedError()
    # Methods with default implementations that may not or may not have very
    # good performance in specific instance classes.
    def drop(self, index=-1):
        """Returns a copy of the persistent sequence with the value at the given
        index dropped.

        If the index is too large or too small for the size of the sequence,
        then the sequence is returned as-is. Non-integer indices will still
        result in raised exceptions."""
        n = len(self)
        if index < -n or index >= n:
            return self
        else:
            return self.delete(index)
    def pop(self, index=-1):
        """Returns tuple of the value at the given index and a copy of the
        persistent sequence with the item at that index removed (default index:
        last)."""
        el = self[index]
        return (el, self.delete(index))
    def remove(self, value):
        """Returns a new list with the first occurence of value removed.

        Raises ValueError if the value is not present.
        """
        ii = self.index(value)
        return self.delete(ii)
    def sort(self, key=None, reverse=False):
        """Returns a sorted copy of the given persistent sequence."""
        t = self.clear().transient()
        for k in sorted(self, key=key, reverse=reverse):
            t.append(k)
        return type(self)(t)
    def reverse(self):
        """Returns a plist that is a reversed copy."""
        t = self.clear().transient()
        for el in self.__reversed__():
            t.append(el)
        return type(self)(t)
    def __str__(self):
        # We have a max length of 60 characters, not counting the delimiters.
        return f"[|{seqstr(self, maxlen=60)}|]"
    def __repr__(self):
        #s = repr(list(self))
        #return f"[|{s[1:-1]}|]"
        return f"[|{seqstr(self)}|]"
    _eq_types = ()
    def __eq__(self, other):
        if other is self:
            return True
        elif not isinstance(other, PersistentSequence._eq_types):
            return False
        else:
            return seqcmp(self, other) == 0
    def __lt__(self, other):
        if not isinstance(other, (list, PersistentSequence, TransientSequence)):
            raise TypeError(f"'<' not supported between instances of"
                            f" '{type(self)}' and '{type(other)}'")
        return seqcmp(self, other) < 0
    def __le__(self, other):
        if not isinstance(other, (list, PersistentSequence, TransientSequence)):
            raise TypeError(f"'<=' not supported between instances of"
                            f" '{type(self)}' and '{type(other)}'")
        return seqcmp(self, other) <= 0
    def __gt__(self, other):
        if not isinstance(other, (list, PersistentSequence, TransientSequence)):
            raise TypeError(f"'>' not supported between instances of"
                            f" '{type(self)}' and '{type(other)}'")
        return seqcmp(self, other) > 0
    def __ge__(self, other):
        if not isinstance(other, (list, PersistentSequence, TransientSequence)):
            raise TypeError(f"'>=' not supported between instances of"
                            f" '{type(self)}' and '{type(other)}'")
        return seqcmp(self, other) >= 0
    def __hash__(self):
        return hash(tuple(self)) + 1
    def __contains__(self, value):
        for el in self:
            if el == value:
                return True
        return False
    def __reversed__(self):
        n = len(self)
        return map(self.__getitem__, range(n - 1, -1, -1))
    def count(self, value):
        """Returns the number of occurences of value."""
        n = 0
        for (k,obj) in self._phamt:
            if obj == value: n += 1
        return n
    def extend(self, iterable):
        """Return a new persistent sequence with the iterables appended."""
        t = self.transient()
        for el in iterable:
            t.append(el)
        return type(self)(t)
    def index(self, value, start=0, stop=None):
        """Returns first index of value.

        Raises ValueError if value is not present.
        """
        if stop is None:
            stop = len(self)
        if start == 0:
            for (ii,val) in enumerate(self):
                if ii >= stop:
                    break
                elif val == value:
                    return ii
        else:
            for (ii,val) in enumerate(self):
                if ii >= stop:
                    break
                elif ii < start:
                    pass
                elif val == value:
                    return ii
        raise ValueError(f"{value} is not in {type(self)}")
    # We include a hash function; though this should be overwritten to cache the
    # hash-code due to it's slowness.
    def __add__(self, obj):
        if not isinstance(obj, (list, PersistentSequence)):
            msg = (f"unsuppoorted operand type for +:"
                   f" '{type(obj)}' and '{type(self)}'")
            raise TypeError(msg)
        elif len(obj) == 0:
            return self
        elif len(self) == 0 and isinstance(obj, PersistentSequence):
            return obj
        else:
            return self.extend(obj)
    def __radd__(self, obj):
        if isinstance(obj, PersistentSequence):
            return NotImplemented
        elif not isinstance(obj, list):
            msg = (f"unsuppoorted operand type for +:"
                   f" '{type(self)}' and '{type(obj)}'")
            raise TypeError(msg)
        elif len(obj) == 0:
            return self
        elif len(self) == 0 and isinstance(obj, PersistentSequence):
            return obj
        else:
            t = self.transient()
            for el in reversed(obj):
                t.prepend(el)
            return type(self)(t)
    def __mul__(self, value):
        if not isinstance(value, Integral):
            msg = f"can't multiply sequence by non-int of type '{type(value)}'"
            raise ValueError(msg)
        reps = int(value)
        if reps < 0:
            reps = 0
        if reps == 0:
            return self.clear()
        elif reps == 1:
            return self
        t = self.transient()
        for r in range(reps - 1):
            for el in self:
                t.append(el)
        return type(self)(t)
    def __rmul__(self, value):
        return self.__mul__(value)
    # For pickling:
    def __reduce__(self):
        return (self.__new__, (type(self), list(self),))
    def __json__(self):
        from json import dumps
        return dumps(list(self))


#===============================================================================
# TransientSequence

class TransientSequence(Transient, MutableSequence):
    """All the operations on a transient sequence.

    Transient sequences are mutable sequences (i.e., objects that inherit from
    `collections.abc.MutableSequence`), but they differ from other sequences in
    that they support efficient conversion to and from a paired persistent
    datatype.

    The following abstract methods must be implemented; if these methods are
    inherited from a superclass of `TransientSequence`, that class is noted in
    parentheses.
     * `clear()`
     * `persistent()` (`Transient`)
     * `__iter__` (`Iterable`)
     * `__len__` (`Sized`)
     * `__getitem__` (`Sequence`)
     * `__setitem__(index, object)` (`MutableSequence`)
     * `__delitem__(index)` (`MutableSequence`)
     * `append(object)` (`MutableSequence`)
     * `prepend(object)` (`MutableSequence`)
     * `insert(index, object)` (`MutableSequence`)

    Additionally, `TransientSequence` includes default implementations of the
    following methods, which may or may not be optimal for any particular
    base-class.
     * `pop(index=-1)`
     * `remove(value)`
     * `sort()`
     * `reverse()`
     * `__str__` (`object`)
     * `__repr__` (`object`)
     * `__eq__` (`object`)
     * `__contains__` (`Container`)
     * `__reversed__` (`Reversible`)
     * `count` (`Sequence`)
     * `extend(iterable)` (`Sequence`)
     * `index(value)` (`Sequence`)
     * `copy()` (`Transient`)
     * `__add__`
     * `__radd__`
     * `__iadd__`
     * `__mul__`
     * `__rmul__`
     * `__imul__`
     * `__reduce__` (for pickling)
     * `__json__` (for `json_fix` module)
    """
    def pop(self, index=-1):
        """Remove and return item at index (default last).

        Raises IndexError if list is empty or index is out of range."""
        el = self[index]
        del self[index]
        return el
    def remove(self, value):
        """Removes the first occurence of value.

        Raises ValueError if the value is not present.
        """
        ii = self.index(value)
        del self[ii]
    def sort(self, key=None, reverse=False):
        """Sort the tlist in ascending order and return None.

        The sort is in-place (i.e. the tlist itself is modified) and stable
        (i.e. the order of two equal elements is maintained).

        If a key function is given, apply it once to each list item and sort
        them, ascending or descending, according to their function values.

        The reverse flag can be set to sort in descending order.
        """
        for (ii,el) in enumerate(sorted(self, key=key, reverse=reverse)):
            t[ii] = el
    def count(self, value):
        """Returns the number of occurences of value."""
        n = 0
        for obj in self:
            if obj == value: n += 1
        return n
    def extend(self, iterable):
        """Extends tlist by appending elements from the iterable."""
        for val in iterable:
            self.append(val)
    def index(self, value, start=0, stop=None):
        """Returns the first index of value.

        Raises ValueError if value is not present.
        """
        for (ii,el) in enumerate(self):
            if value == el:
                return ii
        raise ValueError(f'{value} is not in {type(self)}')
    def reverse(self):
        """Reverses *IN PLACE*."""
        n = len(self)
        for k in range(n//2):
            kk = n - k
            tmp = self[k]
            self[k] = self[kk]
            self[kk] = tmp
    def __str__(self):
        # We have a max length of 60 characters, not counting the delimiters.
        return f"[<{seqstr(self, maxlen=60)}>]"
    def __repr__(self):
        return f"[<{seqstr(self)}>]"
    _eq_types = ()
    def __eq__(self, other):
        if other is self:
            return True
        elif not isinstance(other, TransientSequence._eq_types):
            return False
        elif len(self) != len(other):
            return False
        else:
            return seqcmp(self, other) == 0
    def __lt__(self, other):
        if not isinstance(other, (list, PersistentSequence, TransientSequence)):
            raise TypeError(f"'<' not supported between instances of"
                            f" '{type(self)}' and '{type(other)}'")
        return seqcmp(self, other) < 0
    def __le__(self, other):
        if not isinstance(other, (list, PersistentSequence, TransientSequence)):
            raise TypeError(f"'<=' not supported between instances of"
                            f" '{type(self)}' and '{type(other)}'")
        return seqcmp(self, other) <= 0
    def __gt__(self, other):
        if not isinstance(other, (list, PersistentSequence, TransientSequence)):
            raise TypeError(f"'>' not supported between instances of"
                            f" '{type(self)}' and '{type(other)}'")
        return seqcmp(self, other) > 0
    def __ge__(self, other):
        if not isinstance(other, (list, PersistentSequence, TransientSequence)):
            raise TypeError(f"'>=' not supported between instances of"
                            f" '{type(self)}' and '{type(other)}'")
        return seqcmp(self, other) >= 0
    def __contains__(self, value):
        for el in self:
            if el == value:
                return True
        return False
    def __reversed__(self):
        n = len(self)
        return map(self.__getitem__, range(n - 1, -1, -1))
    def count(self, value):
        """Returns the number of occurences of value."""
        n = 0
        for (k,obj) in self._phamt:
            if obj == value: n += 1
        return n
    def extend(self, iterable):
        """Return a new persistent sequence with the iterables appended."""
        for el in iterable:
            self.append(el)
    def index(self, value, start=0, stop=None):
        """Returns first index of value.

        Raises ValueError if value is not present.
        """
        if start == 0:
            for (ii,val) in enumerate(self):
                if ii >= stop:
                    break
                elif val == value:
                    return ii
        else:
            for (ii,val) in enumerate(self):
                if ii >= stop:
                    break
                elif ii < start:
                    pass
                elif val == value:
                    return ii
        raise ValueError(f"{value} is not in {type(self)}")
    def __iadd__(self, obj):
        if not isinstance(obj, (list, PersistentSequence, TransientSequence)):
            msg = (f"unsuppoorted operand type for +=:"
                   f" '{type(obj)}' and '{type(self)}'")
            raise TypeError(msg)
        elif len(obj) == 0:
            pass
        elif obj is self:
            self.extend(self.persistent())
        else:
            self.extend(obj)
        return self
    def __add__(self, obj):
        if not isinstance(obj, (list, PersistentSequence, TransientSequence)):
            return NotImplemented
        t = self.copy()
        t += obj
        return t
    def __radd__(self, obj):
        if not isinstance(obj, (list, PersistentSequence, TransientSequence)):
            msg = (f"unsuppoorted operand type for +:"
                   f" '{type(self)}' and '{type(obj)}'")
            raise TypeError(msg)
        elif len(obj) == 0:
            return self.copy()
        else:
            t = self.copy()
            for el in reversed(obj):
                t.prepend(el)
            return r
    def __imul__(self, value):
        if not isinstance(value, Integral):
            msg = f"can't multiply sequence by non-int of type '{type(value)}'"
            raise ValueError(msg)
        reps = int(value)
        if reps < 0:
            reps = 0
        if reps == 0:
            self.clear()
        elif reps > 1:
            n = len(self)
            for r in range(reps - 1):
                for (ii,el) in enumerate(self):
                    if ii > n:
                        break
                    else:
                        self.append(el)
        return self
    def __mul__(self, value):
        t = self.copy()
        t *= value
        return t
    def __rmul__(self, value):
        return self.__mul__(value)
    def copy(self):
        """Returns a copy of the given transient sequence."""
        return self.persistent().transient()
    # For pickling.
    def __reduce__(self):
        return (self.__new__, (type(self), list(self),))
    def __json__(self):
        from json import dumps
        return dumps(list(self))    

# Setup the _eq_types, which decides what types can be considered equal.
PersistentSequence._eq_types = (list, PersistentSequence, TransientSequence)
TransientSequence._eq_types = (list, PersistentSequence, TransientSequence)



# -*- coding: utf-8 -*-
################################################################################
# pcollections/util/__init__.py
# Initialization file for the pcollections utilities.
# By Noah C. Benson

"""Utility functions for persistent and transient collections.
"""

from ._core import (setcmp, seqcmp, seqstr)



# -*- coding: utf-8 -*-
################################################################################
# pcollections/_set.py
# The persistent set type for Python.
# By Noah C. Benson

from phamt import (PHAMT,THAMT)

from .abc  import (PersistentSet, TransientSet)
from .util import (setcmp)


#===============================================================================
# pset
# The persistent set type.

class pset(PersistentSet):
    """A persistent set type similar to `set` or `frozenset`.

    `pset()` returns the empty `pset`.

    `pset(iterable)` returns a `pset` containing the elements in `iterable`.

    Unlike `set`, `pset` is immutable; unlike `frozenset`, `pset` supports the
    efficient creation of new `pset` objects that are identical but with small
    changes such as the inclusion of an additional element.
    """
    empty = None
    @classmethod
    def _new(cls, els, idx, top):
        new_pset = super(pset, cls).__new__(cls)
        object.__setattr__(new_pset, '_els', els)
        object.__setattr__(new_pset, '_idx', idx)
        object.__setattr__(new_pset, '_top', top)
        object.__setattr__(new_pset, '_hashcode', None)
        return new_pset
    __slots__ = ("_els", "_idx", "_top", "_hashcode")
    def __new__(cls, *args, **kw):
        if len(kw) > 0:
            raise TypeError("pset() takes no keyword arguments")
        n = len(args)
        if n == 1:
            arg = args[0]
        elif n == 0:
            return pset.empty
        else:
            raise TypeError(f"pset expects at most 1 argument, got {n}")
        # If arg is a tset, this is a special case.
        if isinstance(arg, tset):
            if len(arg) == 0:
                return cls.empty
            else:
                return cls._new(arg._els.persistent(),
                                arg._idx.persistent(),
                                arg._top)
        # If it's a pset, we can just return it as-is.
        if isinstance(arg, pset):
            return arg
        # For anything else, however, we just route this through tset.
        return tset(arg).persistent()
    def __len__(self):
        return len(self._els)
    def __contains__(self, el):
        h = hash(el)
        ii = self._idx.get(h, None)
        while ii is not None:
            (x,ii) = self._els[ii]
            if el == x:
                return True
        return False
    def __iter__(self):
        return map(lambda u: u[1][0], self._els)
    def __hash__(self):
        if self._hashcode is None:
            h = PersistentSet.__hash__(self)
            object.__setattr__(self, '_hashcode', h)
        return self._hashcode
    def transient(self):
        """Returns a transient copy of the set in constant time."""
        return tset._new(THAMT(self._els), THAMT(self._idx), self._top, self)
    def add(self, obj):
        """Returns a copy of the pset that includes the given object."""
        # Get the hash and initial index (if there is one).
        h = hash(obj)
        ii = self._idx.get(h, None)
        if ii is None:
            # The object's hash is not here yet, so we can append to els and
            # insert it into idx.
            new_els = self._els.assoc(self._top, (obj, None))
            new_idx = self._idx.assoc(h, self._top)
        else:
            # First make sure it's not already in the set.
            ii_prev = None
            while ii is not None:
                (x,ii_next) = self._els[ii]
                if obj == x:
                    return self
                ii_prev = ii
                x_prev = x
                ii = ii_next
            # If we reach this point, we can add the object to the end of the
            # list.
            new_els = self._els.assoc(self._top, (obj, None))
            new_els = new_els.assoc(ii_prev, (x_prev, self._top))
        return self._new(new_els, new_idx, self._top + 1)
    def discard(self, obj):
        """Returns a copy of the pset that does not include the given object.

        If the element is not a member, `discard` returns the original pset.
        """
        # Get the hash and initial index (if there is one).
        h = hash(obj)
        ii = self._idx.get(h, None)
        # First make sure it's not already in the set.
        ii_prev = None
        x_prev = None
        while ii is not None:
            (x,ii_next) = self._els[ii]
            if obj == x:
                # We remove this object!
                if ii_prev is None:
                    # We're removing from the front of the list.
                    if ii_next is None:
                        new_idx = self._idx.dissoc(h)
                    else:
                        new_idx = self._idx.assoc(h, ii_next)
                    new_els = self._els
                else:
                    # We're removing from the end or the middle.
                    new_idx = self._idx
                    new_els = self._els.assoc(ii_prev, (x_prev, ii_next))
                new_els = new_els.dissoc(ii)
                return self._new(new_els, new_idx, self._top)
            ii_prev = ii
            x_prev = x
            ii = ii_next
        # If we reach this point, then obj isn't in the set, so we just return
        # self unchanged.
        return self
    def clear(self):
        """Returns the empty pset."""
        return pset.empty
# Make the empty pset.
pset.empty = pset._new(PHAMT.empty, PHAMT.empty, 0)


#===============================================================================
# tset
# The transient set type.

class tset(TransientSet):
    """A transient set type similar to `set` for mutating persistent sets.

    `tset()` returns the empty `tset`.

    `tset(iterable)` returns a `tset` containing the elements in `iterable`.

    `tset(p)` efficiently returns a transient copy of the persistent set `p`.

    Unlike `set`, `tset` is immutable; unlike `frozenset`, `tset` supports the
    efficient creation of new `tset` objects that are identical but with small
    changes such as the inclusion of an additional element.
    """
    @classmethod
    def _new(cls, els, idx, top, orig=None):
        new_tset = super(tset, cls).__new__(cls)
        object.__setattr__(new_tset, '_els', els)
        object.__setattr__(new_tset, '_idx', idx)
        object.__setattr__(new_tset, '_top', top)
        object.__setattr__(new_tset, '_orig', orig)
        return new_tset
    @classmethod
    def empty(cls):
        """Returns an empty tset."""
        return cls._new(THAMT(PHAMT.empty), THAMT(PHAMT.empty), 0)
    __slots__ = ("_els", "_idx", "_top", "_orig")
    def __new__(cls, *args, **kw):
        if len(kw) > 0:
            raise TypeError("tset() takes no keyword arguments")
        n = len(args)
        if n == 1:
            pass
        elif n == 0:
            return cls.empty()
        else:
            raise TypeError(f"tset expects at most 1 argument, got {n}")
        arg = args[0]
        # If arg is a pset, this is a special case.
        if isinstance(arg, pset):
            return cls._new(THAMT(arg._els), THAMT(arg._idx), arg._top)
        # For anything else, however, we just build up.
        t = cls.empty()
        t.addall(arg)
        return t
    def __len__(self):
        return len(self._els)
    def __contains__(self, el):
        h = hash(el)
        ii = self._idx.get(h, None)
        while ii is not None:
            (x,ii) = self._els[ii]
            if el == x:
                return True
        return False
    def __iter__(self):
        return map(lambda u: u[1][0], self._els)
    def add(self, obj):
        """Returns a copy of the tset that includes the given object."""
        # Get the hash and initial index (if there is one).
        h = hash(obj)
        ii_first = self._idx.get(h, None)
        if ii_first is None:
            # The object's hash is not here yet, so we can append to els and
            # insert it into idx.
            self._els[self._top] = (obj, None)
            self._idx[h] = self._top
        else:
            # First make sure it's not already in the set.
            ii = ii_first
            ii_prev = None
            while ii is not None:
                (x,ii_next) = self._els[ii]
                if obj == x:
                    return None
                ii_prev = ii
                x_prev = x
                ii = ii_next
            # If we reach this point, we can add the object to the end of the
            # list.
            self._els[self._top] = (obj, None)
            self._els[ii_prev] = (x_prev, self._top)
        object.__setattr__(self, '_top', self._top + 1)
        object.__setattr__(self, '_orig', None)
    def discard(self, obj):
        """Discards the given object from the set.

        If the element is not a member, `discard` simply returns.
        """
        # Get the hash and initial index (if there is one).
        h = hash(obj)
        ii = self._idx.get(h, None)
        # First make sure it's not already in the set.
        ii_prev = None
        x_prev = None
        while ii is not None:
            (x,ii_next) = self._els[ii]
            if obj == x:
                # We remove this object!
                if ii_prev is None:
                    # We're removing from the front of the list.
                    if ii_next is None:
                        del self._idx[h]
                    else:
                        self._idx[h] = ii_next
                    del self._els[ii]
                else:
                    # We're removing from the end or the middle.
                    del self._els[ii]
                    self._els[ii_prev] = (x_prev, ii_next)
                object.__setattr__(self, '_orig', None)
                return None
            ii_prev = ii
            x_prev = x
            ii = ii_next
        # If we reach this point, then obj isn't in the set, so we just return.
        return None
    def clear(self):
        """Clears the tset."""
        object.__setattr__(self, '_els', THAMT(PHAMT.empty))
        object.__setattr__(self, '_idx', THAMT(PHAMT.empty))
        object.__setattr__(self, '_top', 0)
        object.__setattr__(self, '_orig', None)
    def persistent(self):
        """Efficiently returns a persistent set that is a copy of the tset."""
        if len(self) == 0:
            return pset.empty
        elif self._orig is None:
            return pset._new(self._els.persistent(),
                             self._idx.persistent(),
                             self._top)
        else:
            return self._orig





#! /usr/bin/env python
################################################################################

import os
from setuptools import setup

# Get the version from __init__.py
with open(os.path.join('pcollections', '__init__.py'), 'rt') as fl:
    lns = fl.readlines()
version = next(ln for ln in lns if "__version__ = " in ln)
version = version.split('"')[1]

setup(
    name='pcollections',
    version=version,
    description='Persistent Collections for Python',
    keywords='persistent immutable functional', 
    author='Noah C. Benson',
    author_email='nben@uw.edu',
    url='https://github.com/noahbenson/pcollections/',
    download_url='https://github.com/noahbenson/pcollections/',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: Unix',
        'Operating System :: MacOS',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules'],
    packages=['pcollections',
              'pcollections.abc',
              'pcollections.util',
              'pcollections.test'],
    package_data={'': ['LICENSE.txt']},
    zip_safe=False,
    include_package_data=True,
    install_requires=['phamt >= 0.1.7'])



# -*- coding: utf-8 -*-
################################################################################
# pcollections/test/_list.py
# Declaration of tests for the plist and tlist types.
# By Noah C. Benson

from random import randint
from unittest import TestCase

from .._list import (plist, tlist)
from .._lazy import (lazy, llist, tllist)

class TestPList(TestCase):
    """Tests for the `plist` and `tlist` classes.

    This both runs a number of simple tests of the `plist` API and a series of
    randomized tests in which `plist` must match the behavior of Python's
    native `list` type.
    """
    def test_api(self):
        """Tests/demonstrates the basic plist API."""
        # An empty plist can be created with `plist()`.
        e = plist()
        self.assertEqual(len(e), 0)
        self.assertIsInstance(e, plist)
        # This is identical to the plist.empty object.
        self.assertIs(e, plist.empty)
        # A plist can also be created by passing an iterable to the type.
        p1 = plist(range(10))
        self.assertEqual(len(p1), 10)
        # Element access is like with the list type.
        self.assertEqual(p1[0], 0)
        self.assertEqual(p1[9], 9)
        self.assertEqual(p1[-2], 8)
        # Slice access is supported (slices are plists)
        self.assertEqual(p1[0:10:2], [0,2,4,6,8])
        self.assertIsInstance(p1[:-1], plist)
        # plists can be equal to each other and to other lists (but, like with
        # list, not to other sequences like tuples or strings).
        l1 = list(range(10))
        self.assertEqual(l1, p1)
        self.assertEqual(p1, l1)
        self.assertEqual(p1, plist(range(10)))
        self.assertNotEqual(p1, tuple(p1))
        self.assertNotEqual(plist.empty, '')
        # plists can perform ordering comparisons like list as well.
        self.assertLessEqual(p1[:4], p1)
        self.assertGreater(p1[4:], p1)
        self.assertLess(p1[:4], p1[6:])
        # Clearing a plist always yields the empty plist.
        self.assertIs(p1.clear(), plist.empty)
        # Copying a plist always just returns the plist (it is immutable).
        self.assertIs(p1.copy(), p1)
        # plists are hashable.
        self.assertIsInstance(hash(p1), int)
        # plists contain values.
        self.assertIn(8, p1)
        self.assertNotIn(11, p1)
        # They can be iterated.
        self.assertEqual(p1, list(iter(p1)))
        # They have lengths.
        self.assertEqual(len(p1), 10)
        # They can count and index items.
        self.assertEqual(p1.count(4), 1)
        self.assertEqual(p1.count(-4), 0)
        self.assertEqual(plist([1,2,3,3,4,5,3,3,6]).count(3), 4)
        self.assertEqual(p1.index(5), 5)
        # The reverse method does not mutate it in-place; instead it returns
        # a reversed plist. This is equivalent to the __reversed__ method.
        self.assertEqual(p1.reverse(), list(reversed(p1)))
        self.assertIsNot(p1.reverse(), p1)
        self.assertIsInstance(p1.reverse(), plist)
        # The sort method also returns a sorted plist.
        self.assertEqual(p1.sort(), p1)
        self.assertEqual(p1.reverse().sort(), p1)
        # Mutating elements is done by creating a copy with the desired mutation
        # using the set method. This leaves the original object unchanged.
        p2 = p1.set(0, 10)
        self.assertEqual(p2[0], 10)
        self.assertEqual(p1[0], 0)
        self.assertGreater(p2, p1)
        self.assertEqual(p1[1:], p2[1:])
        # Deleting elements can be done using the delete or drop methods.
        self.assertEqual(p2[:4].delete(1), [10, 2, 3])
        self.assertEqual(p2[:4].drop(2), [10, 1, 3])
        # Drop and delete only differ in their behavior when the index is
        # invalid.
        with self.assertRaises(IndexError):
            p2.delete(20)
        self.assertEqual(p2, p2.drop(20))
        # Items can be appended and prepended to the lists.
        p3 = plist([1,2,3])
        self.assertEqual(p3.append(4), [1,2,3,4])
        self.assertEqual(p3.prepend(0), [0,1,2,3])
        # The extend method also appends sequences, which is basically the same
        # as list + sequence.
        self.assertEqual(p3.extend([4,5,6]), [1,2,3,4,5,6])
        self.assertEqual(p3.extend([4,5,6]), p3 + [4,5,6])
        self.assertEqual([-1,0] + p3, [-1,0,1,2,3])
        # plists can also be multiplied.
        self.assertEqual(plist([1]) * 5, [1,1,1,1,1])
        self.assertEqual(5 * plist([1]), [1,1,1,1,1])
        # The pop method may be used to extract an item.
        self.assertEqual(p3.pop(1), (2, [1,3]))
        # Finally, the insert method can be used to insert items.
        self.assertEqual(p3.insert(1, 10), [1, 10, 2, 3])
    def test_immutable(self):
        """Ensures that `plist` throws the right errors when one mutates it."""
        l = plist(range(10))
        # Cannot set-item.
        with self.assertRaises(TypeError):
            l[0] = 10
        # Cannot del-item.
        with self.assertRaises(TypeError):
            del l[0]
        # Cannot set-attr.
        with self.assertRaises(TypeError):
            l._start = -10
    def test_mul(self):
        "Tests the plist multiplication operator."
        l = plist(range(10))
        # Zero * plist is an empty plist.
        self.assertIs(l * 0, plist.empty)
        self.assertIs(0 * l, plist.empty)
        # List * 2 doubles the list.
        ll = l * 2
        self.assertEqual(l, ll[:10])
        self.assertEqual(l, ll[10:])
        self.assertEqual(ll[:10], l)
        self.assertEqual(ll[10:], l)
        # Cannot multiply by a non-integer
        for notint in ['x', [1], 5.5]:
            with self.assertRaises(ValueError):
                u = l * notint
            with self.assertRaises(ValueError):
                u = notint * l
    def test_add(self):
        "Tests the plist addition operator."
        l = plist(range(10))
        # A plist plus an empty list is the same plist.
        self.assertIs(l, l + [])
        self.assertIs(l, l + plist.empty)
        self.assertIs(l, plist.empty + l)
        # If the list comes first, the return value is a list.
        self.assertEqual(list(l), [] + l)
        # Doubling a list:
        for ll in (l + l, l + list(l)):
            self.assertEqual(l, ll[:10])
            self.assertEqual(l, ll[10:])
        # Cannot add non-lists to lists.
        for notlist in [5,'5',(5,)]:
            with self.assertRaises(TypeError):
                u = notlist + l
            with self.assertRaises(TypeError):
                u = l + notlist
    def test_random(self):
        "Performs a randomized test on the plist type."
        nops = 100
        valmax = 1000
        p = plist()
        t = tlist()
        l = list()
        for opnum in range(nops):
            op = randint(0, 9)
            el = randint(0, valmax)
            if op < 2:
                #print(f"append {el} at {len(l)}")
                p = p.append(el)
                t.append(el)
                l.append(el)
            elif op < 4:
                #print(f"prepend {el} at {len(l)}")
                p = p.prepend(el)
                t.prepend(el)
                l = [el] + l
            elif op < 6:
                ii = randint(0, len(l))
                #print(f"insert ({ii}, {el}) at {len(l)}")
                p = p.insert(ii, el)
                t.insert(ii, el)
                l.insert(ii, el)
            elif op == 6:
                if len(p) > 0:
                    lel = l.pop()
                    #print(f"pop -1 {lel} at {len(l)}")
                    (pel, p) = p.pop()
                    tel = t.pop()
                    self.assertEqual(pel, tel)
                    self.assertEqual(pel, lel)
            elif op == 7:
                if len(p) > 0:
                    lel = l.pop(0)
                    #print(f"pop 0 {lel} at {len(l)}")
                    (pel, p) = p.pop(0)
                    tel = t.pop(0)
                    self.assertEqual(pel, tel)
                    self.assertEqual(pel, lel)
            elif op == 8:
                if len(p) > 0:
                    ii = randint(0, len(p) - 1)
                    lel = l.pop(ii)
                    #print(f"pop {ii} {lel} at {len(l) + 1}")
                    (pel, p) = p.pop(ii)
                    tel = t.pop(ii)
                    self.assertEqual(pel, tel)
                    self.assertEqual(pel, lel)
            elif op == 9:
                if len(p) > 0:
                    ii = randint(0, len(p) - 1)
                    #print(f"set {ii} to {el} at {len(l)}")
                    p = p.set(ii, el)
                    t[ii] = el
                    l[ii] = el
            #print('   -', repr(p), ' [', p._start, ']')
            #print('   -', repr(t), ' [', t._start, ']')
            #print('   -', repr(l))
            self.assertEqual(p, t)
            self.assertEqual(p, l)
            self.assertEqual(l, t)
            if randint(0, 20) == 0:
                tmp = t.persistent()
                self.assertEqual(tmp, p)
                self.assertEqual(tmp, t)

class TestLList(TestCase):
    """Tests for the `llist` and `lazy` classes.

    This runs a number of simple tests of the `llist` API; however, because the
    `llist` type shares most of its logic with the `plist` type, the randomized
    tests are not included.
    """
    def test_api(self):
        """Tests/demonstrates the basic llist API."""
        # An empty plist can be created with `llist()`.
        e = llist()
        self.assertEqual(len(e), 0)
        self.assertIsInstance(e, llist)
        self.assertIsInstance(e, plist)
        # This is identical to the llist.empty object.
        self.assertIs(e, llist.empty)
        # ... but it is not identical to the plist.empty object.
        self.assertIsNot(e, plist.empty)
        # A plist can also be created by passing an iterable to the type.
        p1 = llist(range(10))
        self.assertEqual(len(p1), 10)
        # Element access is like with the list type.
        self.assertEqual(p1[0], 0)
        self.assertEqual(p1[9], 9)
        self.assertEqual(p1[-2], 8)
        # Slice access is supported (slices are llists)
        self.assertEqual(p1[0:10:2], [0,2,4,6,8])
        self.assertIsInstance(p1[:-1], llist)
        # llists can be equal to each other and to other lists (but, like with
        # list, not to other sequences like tuples or strings).
        l1 = list(range(10))
        self.assertEqual(l1, p1)
        self.assertEqual(p1, l1)
        self.assertEqual(p1, llist(range(10)))
        self.assertEqual(p1, plist(range(10)))
        self.assertNotEqual(p1, tuple(p1))
        self.assertNotEqual(llist.empty, '')
        # llists can perform ordering comparisons like list as well.
        self.assertLessEqual(p1[:4], p1)
        self.assertGreater(p1[4:], p1)
        self.assertLess(p1[:4], p1[6:])
        # Clearing an llist always yields the empty llist.
        self.assertIs(p1.clear(), llist.empty)
        # Copying an llist always just returns the plist (it is immutable).
        self.assertIs(p1.copy(), p1)
        # llists are hashable as long as their values are hashable.
        self.assertIsInstance(hash(p1), int)
        # llists contain values.
        self.assertIn(8, p1)
        self.assertNotIn(11, p1)
        # They can be iterated.
        self.assertEqual(p1, list(iter(p1)))
        # They have lengths.
        self.assertEqual(len(p1), 10)
        # They can count and index items.
        self.assertEqual(p1.count(4), 1)
        self.assertEqual(p1.count(-4), 0)
        self.assertEqual(llist([1,2,3,3,4,5,3,3,6]).count(3), 4)
        self.assertEqual(p1.index(5), 5)
        # The reverse method does not mutate it in-place; instead it returns
        # a reversed llist. This is equivalent to the __reversed__ method.
        self.assertEqual(p1.reverse(), list(reversed(p1)))
        self.assertIsNot(p1.reverse(), p1)
        self.assertIsInstance(p1.reverse(), llist)
        # The sort method also returns a sorted llist.
        self.assertEqual(p1.sort(), p1)
        self.assertEqual(p1.reverse().sort(), p1)
        # Mutating elements is done by creating a copy with the desired mutation
        # using the set method. This leaves the original object unchanged.
        p2 = p1.set(0, 10)
        self.assertEqual(p2[0], 10)
        self.assertEqual(p1[0], 0)
        self.assertGreater(p2, p1)
        self.assertEqual(p1[1:], p2[1:])
        # Deleting elements can be done using the delete or drop methods.
        self.assertEqual(p2[:4].delete(1), [10, 2, 3])
        self.assertEqual(p2[:4].drop(2), [10, 1, 3])
        # Drop and delete only differ in their behavior when the index is
        # invalid.
        with self.assertRaises(IndexError):
            p2.delete(20)
        self.assertEqual(p2, p2.drop(20))
        # Items can be appended and prepended to the lists.
        p3 = llist([1,2,3])
        self.assertEqual(p3.append(4), [1,2,3,4])
        self.assertEqual(p3.prepend(0), [0,1,2,3])
        # The extend method also appends sequences, which is basically the same
        # as list + sequence.
        self.assertEqual(p3.extend([4,5,6]), [1,2,3,4,5,6])
        self.assertEqual(p3.extend([4,5,6]), p3 + [4,5,6])
        self.assertEqual([-1,0] + p3, [-1,0,1,2,3])
        # llists can also be multiplied.
        self.assertEqual(llist([1]) * 5, [1,1,1,1,1])
        self.assertEqual(5 * llist([1]), [1,1,1,1,1])
        # The pop method may be used to extract an item.
        self.assertEqual(p3.pop(1), (2, [1,3]))
        # Finally, the insert method can be used to insert items.
        self.assertEqual(p3.insert(1, 10), [1, 10, 2, 3])
    def test_lazy(self):
        """Tests the lazy aspects of the llist type."""
        # Lazy dictionaries don't evaluate lazy arguments until they are
        # requested, and they only evaluate them once.
        def counter(n):
            counter.count += n
            return counter.count
        counter.count = 0
        p1 = llist([lazy(counter, 1), lazy(counter, 10)])
        self.assertEqual(counter.count, 0)
        self.assertEqual(p1[0], 1)
        self.assertEqual(counter.count, 1)
        # Reaccessing the index will not call counter() again.
        self.assertEqual(p1[0], 1)
        self.assertEqual(counter.count, 1)
        # However, the index 1 has not yet been evaluated, so accessing it will
        # call counter().
        self.assertEqual(p1[1], 11)
        self.assertEqual(counter.count, 11)
        self.assertEqual(p1[1], 11)
        self.assertEqual(counter.count, 11)
        # Iteration also reifies the values.
        counter.count = 0
        p1 = llist([lazy(counter, 1), lazy(counter, 10)])
        self.assertEqual(list(iter(p1)), [1, 11])
        self.assertEqual(p1[0], 1)
        self.assertEqual(p1[1], 11)
        self.assertEqual(counter.count, 11)
        # Converstion via to_pdict preserves the lazy items.
        counter.count = 0
        p1 = llist([lazy(counter, 1), lazy(counter, 10)])
        t1 = p1.to_plist()
        self.assertIsInstance(t1, plist)
        self.assertIsInstance(t1[0], lazy)
        self.assertIsInstance(t1[1], lazy)
        # Conversion to a transient creates a lazy transient.
        t1 = p1.transient()
        self.assertIsInstance(t1, tllist)
        self.assertEqual(t1[0], 1)
        self.assertEqual(t1[1], 11)
        # Note that the new dictionary in this case shares the lazy values with
        # the old dictionary.
        self.assertEqual(p1[0], 1)
        self.assertEqual(p1[1], 11)
        # To extract the lazy object itself, one can use getlazy:
        self.assertIs(p1.getlazy(0), t1.getlazy(0))
        # Equality comparisons depend on the reified, not lazy, values.
        counter.count = 0
        p1 = llist([lazy(counter, 1), lazy(counter, 10)])
        self.assertEqual(p1, [1, 11])
    def test_immutable(self):
        """Ensures that `plist` throws the right errors when one mutates it."""
        l = llist(range(10))
        # Cannot set-item.
        with self.assertRaises(TypeError):
            l[0] = 10
        # Cannot del-item.
        with self.assertRaises(TypeError):
            del l[0]
        # Cannot set-attr.
        with self.assertRaises(TypeError):
            l._start = -10
    def test_mul(self):
        "Tests the plist multiplication operator."
        l = llist(range(10))
        # Zero * plist is an empty plist.
        self.assertIs(l * 0, llist.empty)
        self.assertIs(0 * l, llist.empty)
        # List * 2 doubles the list.
        ll = l * 2
        self.assertEqual(l, ll[:10])
        self.assertEqual(l, ll[10:])
        self.assertEqual(ll[:10], l)
        self.assertEqual(ll[10:], l)
        # Cannot multiply by a non-integer
        for notint in ['x', [1], 5.5]:
            with self.assertRaises(ValueError):
                u = l * notint
            with self.assertRaises(ValueError):
                u = notint * l
    def test_add(self):
        "Tests the plist addition operator."
        l = llist(range(10))
        # A plist plus an empty list is the same plist.
        self.assertIs(l, l + [])
        self.assertIs(l, l + llist.empty)
        self.assertIs(l, llist.empty + l)
        # If the list comes first, the return value is a list.
        self.assertEqual(list(l), [] + l)
        # Doubling a list:
        for ll in (l + l, l + list(l)):
            self.assertEqual(l, ll[:10])
            self.assertEqual(l, ll[10:])
        # Cannot add non-lists to lists.
        for notlist in [5,'5',(5,)]:
            with self.assertRaises(TypeError):
                u = notlist + l
            with self.assertRaises(TypeError):
                u = l + notlist



# -*- coding: utf-8 -*-
################################################################################
# pcollections/test/__main__.py
# Imports the test package so that tests can be run from the command line by
# calling the test module directly.
# By Noah C. Benson

import unittest

from pcollections.test import *

unittest.main()





# -*- coding: utf-8 -*-
################################################################################
# pcollections/util/_core.py
# Implementation of the core pcollections utilities.
# By Noah C. Benson

from collections.abc import (
    Sequence, MutableSequence,
    Set, MutableSet,
    Mapping, MutableMapping,
    Sized, Container
)


#===============================================================================
# Utility Functions

def setcmp(set1, set2):
    """Compares one set to another.

    `setcmp(a, b)` returns -1 if `a` is a subset of `b`, 0 if `a` is equal to
    `b`, and 1 if `a` is a superset of `b`. If `a` and `b` are disjoint, `None`
    is returned. If `a` and `b` have a non-zero intersection but are neither
    subset nor superset, then `Ellipsis` is returned.
    """
    if not isinstance(set1, Set):
        msg = f"unsupported operand type for compare: {type(set1)}"
        raise TypeError(msg)
    if not isinstance(set2, Set):
        msg = f"unsupported operand type for compare: {type(set2)}"
        raise TypeError(msg)
    n1 = len(set1)
    n2 = len(set2)
    (a,na,b,nb) = (set1,n1, set2,n2) if n1 < n2 else (set2,n2, set1,n1)
    isects = 0
    for el in iter(a):
        if el in b:
            isects += 1
    if isects == 0:
        if na == 0:
            if nb == 0:
                return 0 # An empty set is equal to another empty set.
            else:
                return -1 if set1 is a else 1
        else:
            return None # No intersections: the sets are disjoint.
    elif isects < na:
        return Ellipsis # Some but not all elements intersect.
    elif n1 == n2:
        return 0 # All elements of both intersect; the sets are equal.
    elif n1 < n2:
        return -1 # All elements of set1 are in set2, set1 < set2.
    else:
        return 1 # All elements of set2 are in set1, set1 > set2.

def seqcmp(seq1, seq2):
    """Compares one sequence to another.

    `seqcmp(a, b)` returns -1 if `a` sorts before `b`, 0 if `a` is equal to
    `b`, and 1 if `a` sorts after `b`.
    """
    if not isinstance(seq1, Sequence):
        msg = f"unsupported operand type for compare: {type(setq)}"
        raise TypeError(msg)
    if not isinstance(seq2, Sequence):
        msg = f"unsupported operand type for compare: {type(seq2)}"
        raise TypeError(msg)
    for (a,b) in zip(seq1, seq2):
        if a < b:
            return -1
        elif b < a:
            return 1
    n1 = len(seq1)
    n2 = len(seq2)
    if n1 < n2:
        return -1
    elif n1 > n2:
        return 1
    else:
        return 0
    
def seqstr(seq, maxlen=None, sep=", ", tostr=repr):
    """Returns a string representation of the given sequence or iterable.

    Mappings are converted to `"key: value"` strings.

    The option `maxlen` may be set to a positive integer to fill in the string
    with an ellipsis in the case that it goes to long; for exaple,
    `seqstr(range(10), 12)` would return `"0, 1, 2 ..."`.
    """
    if isinstance(seq, Mapping):
        orig_tostr = tostr
        tostr = lambda kv: f"{orig_tostr(kv[0])}: {orig_tostr(kv[1])}"
        seq = seq.items()
    if maxlen is None:
        return sep.join(map(tostr, seq))
    elif not isinstance(maxlen, int) or maxlen < 3:
        raise TypeError(f"maxlen must be an integer >= 3 or None, not {maxlen}")
    seplen = len(sep)
    elllen = 3 + seplen
    parts = []
    N = 0
    for el in seq:
        s = tostr(el)
        parts.append(s)
        n = len(s)
        N = n if N == 0 else N + seplen + n
        if N > maxlen:
            # We've overshot; remove the final element if need-be.
            while len(parts) > 0 and N + elllen > maxlen:
                s = parts.pop()
                N -= len(s) + seplen
            parts.append("...")
            break
    # We successfully added all the elements; we should be able to join them
    # at this point.
    return sep.join(parts)
